# Author: Benoit Da Mota <damota.benoit@gmail.com>
#
# License: BSD 3 clause


def dynamic_import(str_import):
    """Take a string representing a python function or class and import it.

    Parameters:
    -----------
    str_import : str
        the string representing the import (e.g. "sklearn.metrics.f1_score")
    """
    mod, cla = str_import.rsplit('.', 1)
    dyn_import = getattr(__import__(mod, fromlist=[str(cla)]), cla)
    return dyn_import


def _get_step(step):
    """Import class of a given step and return a sklearn.pipeline.Pipeline step
    """
    str_imp = step[1][0]
    kwargs = step[1][1]
    name = step[0]
    return (name, (dynamic_import(str_imp), kwargs))


def construct_pipeline(cfg):
    """Construct the pipeline steps for sklearn.pipeline.Pipeline.

    Construct the pipeline steps for sklearn.pipeline.Pipeline and
    return the steps and the parameter to optimize (e.g. "logit__C" for
    the parameter "C" of the step named "logit").

    Parameters:
    -----------
    cfg : dict,
        method configuration describing the steps of an pipelined estimator
    """
    steps = cfg["steps"]
    est_param = cfg["est_param"]
    pipe = []
    for step in steps:
        pipe.append(_get_step(step))
    return {'steps': pipe}, est_param


def get_score_func(cfg, cv="crossval_score"):
    """Import score function and kwargs from the CV configuration.

    Parameters
    ----------
    cfg : dict,
        configuration dict for cross-validation.

    cv : str, optional(default="crossval_score")
        section of the configuration dict to search for a score function.
    """
    func = dynamic_import(cfg[cv]["funcMetric"][0])
    kwargs = cfg[cv]["funcMetric"][1]
    return func, kwargs


def load_data(cfg):
    """Load data from the data configuration.

    Parameters
    ----------
    cfg : dict,
        configuration dict for data and I/O.
    """
    kwargs = cfg['kwargs']
    dyn_func = dynamic_import(cfg['func'])
    x, y = dyn_func(**kwargs)
    return x, y


def get_grid(cfg, x, y):
    """Load the parameters grid from the CV configuration.

    Parameters
    ----------
    cfg : dict,
        configuration dict for cross-validation.
    x : array, shape (n_samples, n_features)
        features array
    y : array, shape (n_samples, n_targets)
        targets array
    """
    kwargs = cfg["gridSearch"]['parametersGrid'][1]
    dyn_func = dynamic_import(cfg["gridSearch"]['parametersGrid'][0])
    grid = dyn_func(x, y, **kwargs)
    return grid



# Author: Benoit Da Mota <damota.benoit@gmail.com>
#
# License: BSD 3 clause
"""
Build arguments parser for the scripts (mapper, reducers and command builder).
"""
import argparse


def get_map_argparser():
    """Build command line arguments parser for a mapper.

    Arguments parser compatible with the commands builder workflows.
    """
    parser = argparse.ArgumentParser()

    parser.add_argument("crossval",
                        help="JSON file to configure cross validation scheme")
    parser.add_argument("method",
                        help="JSON file to configure the method")
    parser.add_argument("dataset",
                        help="Joblib file with data and folds")
    parser.add_argument("out",
                        help="Filename to output the results")
    parser.add_argument("outer", type=int,
                        help="Outer CV Id")

    parser.add_argument("--inner", type=int,
                        help="Inner CV Id")

    # verbose mode
    parser.add_argument("-v", "--verbose", help="verbose mode",
                        action="store_true")
    return parser


def get_ired_argparser():
    """Build command line arguments parser for an inner reducer.

    Arguments parser compatible with the commands builder workflows.
    """
    parser = argparse.ArgumentParser()

    parser.add_argument("crossval",
                        help="JSON file to configure cross validation scheme")
    parser.add_argument("method",
                        help="JSON file to configure the method")
    parser.add_argument("dataset",
                        help="Joblib file with data and folds")
    parser.add_argument("out",
                        help="Filename to output the results")
    parser.add_argument("in",
                        help="Filename template for input files")
    parser.add_argument("outer", type=int,
                        help="Outer CV Id")

    # verbose mode
    parser.add_argument("-v", "--verbose", help="verbose mode",
                        action="store_true")
    return parser


def get_ored_argparser():
    """Build command line arguments parser for an outer reducer.

    Arguments parser compatible with the commands builder workflows.
    """
    parser = argparse.ArgumentParser()

    parser.add_argument("out",
                        help="Filename to output the results")
    parser.add_argument("in",
                        help="Filename template for input files")

    # verbose mode
    parser.add_argument("-v", "--verbose", help="verbose mode",
                        action="store_true")
    return parser


def get_cmd_builder_argparser():
    """Build command line arguments parser for the commands builder

    Arguments parser for the commands builder workflows.
    """
    parser = argparse.ArgumentParser()
    # data configuration
    parser.add_argument("data",
                        help="JSON file to configure data and I/O")
    # CV configuration
    parser.add_argument("crossval",
                        help="JSON file to configure cross validation scheme")
    # Method configuration
    parser.add_argument("method",
                        help="JSON file to configure the method")
    # local output directory
    parser.add_argument("outputdir",
                        help=("Local directory to store the outputs"
                              " (dataset and configurations)"))

    # output mode {cmd-list, soma-workflow}
    parser.add_argument("-o", "--output-mode",
                        choices=["cmd-list", "soma-workflow"],
                        default="soma-workflow",
                        help="Output mode")
    # verbose mode
    parser.add_argument("-v", "--verbose", help="verbose mode",
                        action="store_true")
    # ignore warnings
    parser.add_argument("--no-warn", help="ignore warnings",
                        action="store_true")
    return parser



#!/usr/bin/env python
# Author: Benoit Da Mota <damota.benoit@gmail.com>
#
# License: BSD 3 clause
"""
Generic inner reducer
"""
import json
import numpy as np

from sklearn.externals import joblib
from sklearn.pipeline import Pipeline

from mempamal.arguments import get_ired_argparser
from mempamal.crossval import get_fold, print_fold
from mempamal.gridsearch import GenericGridSearch
from mempamal.dynamic import construct_pipeline, get_score_func

verbose = False

if __name__ == "__main__":
    # parse command line arguments
    args = get_ired_argparser().parse_args()
    verbose = args.verbose
    if verbose:
        print("=======")
        print(args)
        print("=======")

    # read files
    dataset = joblib.load(args.dataset)
    with open(args.method, 'r') as fd:
        method_cfg = json.load(fd)
    with open(args.crossval, 'r') as fd:
        cv_cfg = json.load(fd)

    # retrieve results from inner folds
    n_inner = dataset["folds"]["n_inner"]
    n_targets = dataset["n_targets"]
    grid = dataset["grid"]
    scores = np.zeros((n_inner, n_targets, grid.size))
    for i in range(n_inner):
        cur_file = (args.__getattribute__("in")).format(inner=i)
        if verbose:
            print("Reading {}".format(cur_file))
        cur_ar = joblib.load(cur_file)
        scores[i] = cur_ar["scores"]
    if verbose:
        print("=======")
    # Parameter selection:
    # mean on the folds and target, i.e. select the best parameters
    # independently of the target (that's one possible strategy for
    # multiple targets)
    ms = np.mean(scores.reshape((-1, grid.size)), axis=0)
    best_param = grid[ms == np.amax(ms)][0]

    # construct folds
    train_index, test_index = get_fold(dataset["folds"], args.outer)
    if verbose:
        print_fold(train_index, test_index)
    X = dataset["X"]
    Y = dataset["Y"]
    X_train = X[train_index]
    Y_train = Y[train_index]
    X_test = X[test_index]
    Y_test = Y[test_index]

    # construct estimator
    est_kwargs, est_param = construct_pipeline(method_cfg)
    score_func, score_kwargs = get_score_func(cv_cfg, cv="gridSearch")
    clf = GenericGridSearch(est=Pipeline,
                            est_param=est_param,
                            params=[best_param],
                            est_kwargs=est_kwargs,
                            score_func=score_func,
                            score_kwargs=score_kwargs)

    # fit/predict/score
    clf.fit(X_train, Y_train)
    Y_pred = clf.predict(X_test)
    if verbose:
        print Y_test
        print Y_pred[0]
    res = {"scores": clf.score(Y_test, Y_pred)[0]}
    print("Best parameters set: {}".format(best_param))
    print("scores: {}".format(res["scores"]))

    # save result
    joblib.dump(res, args.out, compress=1)



# Author: Benoit Da Mota <damota.benoit@gmail.com>
#
# License: BSD 3 clause
"""
Functions relative to configurations.
"""
import json
import os.path as path

from sklearn.pipeline import Pipeline


def _check_conf(cfg, req_keys, cat=""):
    """Generic checking for required keys.

    """
    for k in req_keys:
        try:
            cfg[k]
        except KeyError:
            raise KeyError(("Required key is missing in the "
                            "{} configuration file: {}").format(cat, k))


def _check_cv_conf(cfg):
    """Check the configuration of the cross-validation.

    """
    req_keys = ["stratified", "modelSelection", "crossval_score"]
    _check_conf(cfg, req_keys, cat="crossval")
    req_keys = ["foldsIterator", "funcMetric"]
    _check_conf(cfg["crossval_score"], req_keys, cat="crossval")
    if cfg["modelSelection"]:
        req_keys = ["foldsIterator", "funcMetric", "parametersGrid"]
        _check_conf(cfg, ["gridSearch"], cat="crossval")
        _check_conf(cfg["gridSearch"], req_keys, cat="crossval")


def _check_data_conf(cfg):
    """Check the configuration of the data.

    """
    req_keys = ["func", "kwargs", "in_out_dir"]
    _check_conf(cfg, req_keys, cat="data")


def _check_method_conf(cfg):
    """Check the configuration of the data.

    """
    req_keys = ["mapper", "inner_reducer", "outer_reducer", "steps",
                "est_param"]
    _check_conf(cfg, req_keys, cat="method")


def check_conf(cfg, cat="crossval", verbose=False):
    """Entry point to check a configuration by category.

    Parameters
    ----------
    cfg : dict,
        the configuration to check.
    cat : str in [crossval, method, data], optional (default="crossval")
        category of the configuration to call the good checking method.
    verbose : boolean, optional (default=False)
        verbose mode.
    """
    if cat == "crossval":
        _check_cv_conf(cfg)
    elif cat == "method":
        _check_method_conf(cfg)
    elif cat == "data":
        _check_data_conf(cfg)
    if verbose:
        print(cfg)


def JSONify_estimator(est, est_param,
                      model_selection=True,
                      param_val=None,
                      out=None,
                      path_to_mr=".",
                      mapper="mapper.py",
                      i_red="inner_reducer.py",
                      o_red="outer_reducer.py"):
    """Helper function to jsonify a sklearn.pipeline.Pipeline or an estimator.

    Parameters
    ----------
    est : estimator,
        Estimator to jsonify.
    est_param : str,
        Parameter to optimize.
    model_selection : boolean, optional (default=True)
        Do you plan to perform a model selection (optimization of
        hyper-parameters)?
    param_val : val, optional (default=None)
        If model_selection is False, you should provide a value for
        the est_param.
    out : str, optional (default=None)
        Filename to output the json, if None the json is printed on stdout.
    path_to_mr : str, optional (default=".", i.e. working directory)
        Where to find the mapper and reducers scripts.
    mapper : str, optional (default="mapper.py")
        script for the mapper
    i_red : str, optional (default="inner_reducer.py")
        script for the inner reducer (for model selection)
    o_red : str, optional (default="outer_reducer.py")
        script for the outer reducer
    """
    # introspection of the estimator
    steps = []
    if est.__class__ is Pipeline:
        # if the estimator is a Pipeline, retrieve all steps names,
        # classes and dict
        for n, s in est.steps:
            t = [repr(s.__class__).split("\'")[1], s.__dict__]
            steps.append([n, t])
    else:
        # else the name of the unique step is the name of the
        # estimator class
        cl = repr(est.__class__).split("\'")[1]
        n = cl.split(".")[-1]
        steps.append([n, [cl, est.__dict__]])

    # produce a  method configuration
    conf = {}
    conf["steps"] = steps
    conf["est_param"] = est_param
    if model_selection is False:
        conf[est_param] = param_val
    conf["mapper"] = path.join(path_to_mr, mapper)
    conf["inner_reducer"] = path.join(path_to_mr, i_red)
    conf["outer_reducer"] = path.join(path_to_mr, o_red)

    # output
    if out is None:
        print(json.dumps(conf, indent=2))
    else:
        with open(out, 'w') as fd:
            json.dump(conf, fd, indent=2)



# Author: Benoit Da Mota <damota.benoit@gmail.com>
#
# License: BSD 3 clause
"""
Simple GridSearch for a pipelined estimator (without warm restart).
"""
import numpy as np


class GenericGridSearch(object):
    """Simple GridSearch for a pipelined estimator.

    Note: see sklearn.pipeline.Pipeline
    """

    def __init__(self, est, est_param, params, score_func,
                 est_kwargs=None,
                 score_kwargs=None):
        """

        Parameters
        ----------
        est : estimator,
            Estimator to jsonify.
        est_param : str,
            Parameter to optimize.
        params : array, shape(n_parameters)
            Grid of parameters
        score_func : func,
            scoreng function (e.g. sklearn.metrics.f1_score)
        est_kwargs : dict, optional (default=None)
            keywords arguments for the estimator
        score_kwargs : dict, optional (default=None)
            keywords arguments for the scoring function
        """
        self.params = params
        self.est_param = est_param
        self.est = est
        self.res = {}
        self.score_func = score_func
        if score_kwargs is not None:
            self.score_kwargs = score_kwargs
        else:
            self.score_kwargs = {}
        if est_kwargs is not None:
            self.est_kwargs = est_kwargs
        else:
            self.est_kwargs = {}

    def fit(self, X, y):
        """Fit the estimator on each parameter of the grid.

        Parameters
        ----------
        X : array, shape (n_samples, n_features)
            features array
        y : array, shape (n_samples, n_targets)
            targets array
        """
        for p in self.params:
            param_kwargs = {self.est_param: p}
            steps = self.est_kwargs['steps']
            pipe_steps = []
            for s in steps:
                s_i = (s[1][0])()
                for k, v in (s[1][1]).iteritems():
                    s_i.__setattr__(k, v)
                pipe_steps.append((s[0], s_i))

            self.res[p] = self.est(pipe_steps)
            self.res[p].set_params(**param_kwargs)
            self.res[p].fit(X, y)

    def predict(self, X):
        """Predict the targets from X for each parameter of the grid

        Parameters
        ----------
        X : array, shape (n_samples, n_features)
            features array
        """
        y_pred = []
        for p in self.params:
            y_pred.append(self.res[p].predict(X))
        return np.asarray(y_pred)

    def score(self, y_test, y_pred):
        """Apply the scoring function for each prediction

        Parameters
        ----------
        y_test : array, shape (n_samples, n_targets)
            Real targets values
        y_pred : array, shape (n_parameters, n_samples, n_targets)
            Targets prediction to score.
        """
        scores = []
        for yp in y_pred:
            scores.append(self.score_func(y_test, yp, **self.score_kwargs))
        return np.asarray(scores).T



#!/usr/bin/env python
# Author: Benoit Da Mota <damota.benoit@gmail.com>
#
# License: BSD 3 clause
"""
Generic mapper
"""
import json

from sklearn.externals import joblib
from sklearn.pipeline import Pipeline

from mempamal.arguments import get_map_argparser
from mempamal.crossval import get_fold, print_fold
from mempamal.gridsearch import GenericGridSearch
from mempamal.dynamic import construct_pipeline, get_score_func

verbose = False

if __name__ == "__main__":
    # parse command line arguments
    args = get_map_argparser().parse_args()
    verbose = args.verbose
    if verbose:
        print("=======")
        print(args)
        print("=======")

    # read data and configuration files
    dataset = joblib.load(args.dataset)
    grid = dataset["grid"]
    with open(args.crossval, 'r') as fd:
        cv_cfg = json.load(fd)
    with open(args.method, 'r') as fd:
        method_cfg = json.load(fd)

    # construct folds
    train_index, test_index = get_fold(dataset["folds"],
                                       args.outer, inner=args.inner)
    if verbose:
        print_fold(train_index, test_index)
    X = dataset["X"]
    Y = dataset["Y"]
    X_train = X[train_index]
    Y_train = Y[train_index]
    X_test = X[test_index]
    Y_test = Y[test_index]

    # construct estimator
    est_kwargs, est_param = construct_pipeline(method_cfg)
    which_cv = ("gridSearch" if cv_cfg["modelSelection"]
                else "crossval_score")
    score_func, score_kwargs = get_score_func(cv_cfg, cv=which_cv)
    clf = GenericGridSearch(est=Pipeline,
                            est_param=est_param,
                            params=grid,
                            est_kwargs=est_kwargs,
                            score_func=score_func,
                            score_kwargs=score_kwargs)

    # fit/predict/score
    clf.fit(X_train, Y_train)
    Y_pred = clf.predict(X_test)
    if verbose:
        print Y_test
        print Y_pred
    scores = clf.score(Y_test, Y_pred)
    res = ({"scores": scores} if cv_cfg["modelSelection"]
           else {"scores": scores[0]})

    # save result
    joblib.dump(res, args.out, compress=1)



#!/usr/bin/env python
# Author: Benoit Da Mota <damota.benoit@gmail.com>
#
# License: BSD 3 clause
"""
Generic outer reducer.
"""
from glob import glob

import numpy as np
from sklearn.externals import joblib

from mempamal.arguments import get_ored_argparser

verbose = False

if __name__ == "__main__":
    # parse command line arguments
    args = get_ored_argparser().parse_args()
    verbose = args.verbose
    if verbose:
        print("=======")
        print(args)
        print("=======")

    # retrieve results from (outer) folds
    file_pattern = (args.__getattribute__("in")).format(outer="*")
    list_files = glob(file_pattern)
    scores = []
    for cur_file in list_files:
        if verbose:
            print("Reading {}".format(cur_file))
        cur_ar = joblib.load(cur_file)
        scores.append(cur_ar["scores"])
    raw = np.asarray(scores)

    # summary
    res = {"raw": raw,
           "mean": np.mean(raw, axis=0),
           "median": np.median(raw, axis=0),
           "std": np.std(raw, axis=0)}
    if verbose:
        print("=======")
        print(res)
        print("=======")
    print("Cross-validated score(s):")
    print("  Mean  : %s" % (res['mean']).__str__())
    print("  Median: %s" % (res['median']).__str__())
    print("  Std   : %s" % (res['std']).__str__())

    # save result
    joblib.dump(res, args.out, compress=1)



# Author: Benoit Da Mota <damota.benoit@gmail.com>
#
# License: BSD 3 clause
"""
Example of dataset and how to load its data.
"""
from sklearn.datasets import load_iris


def get_data(**kwargs):
    """Load data and return a x, y.

    Note: it ignores all arguments.
    """
    data = load_iris()
    x = data['data']
    y = data['target']
    return x, y



# Author: Benoit Da Mota <damota.benoit@gmail.com>
#
# License: BSD 3 clause
"""
Functions relative to cross-validation and folds iterator.
"""
import numpy as np

from .dynamic import dynamic_import


def _construct_folds_iterator(y, cfg, key):
    """Dynamically construct a fold iterator from the configuration (cfg, key).

    Instantiate a folds iterator depending on the cross-validation
    configuration. If the folds iterator is statified, it should take the
    targets array y, else it should take only ne number of samples.

    Notes
    -----
    See sklearn.cross_validation for more information about how a
    folds iterator should work

    Parameters
    ----------
    y : array, shape (n_samples, n_targets)
        targets array
    cfg : dict,
        configuration dict for cross-validation.
    key : str,
        key in cfg that provide a key "foldsIterator" and a list
        [class, dict] where 'class' is a string representing the
        folds iterator object to import and 'dict' are the keywords
        arguments.
    """
    cv_cfg = cfg[key]
    dyn_cla = dynamic_import(cv_cfg["foldsIterator"][0])
    kwargs = cv_cfg["foldsIterator"][1]
    if cfg["stratified"]:
        fg = dyn_cla(y, **kwargs)
    else:
        fg = dyn_cla(y.shape[0], **kwargs)
    return fg


def make_folds(y, cfg, verbose=False):
    """Constructs the folds from crossval configuration

    Parameters
    ----------
    y : array, shape (n_samples, n_targets)
        targets array
    cfg : dict,
        configuration dict for cross-validation.
    verbose : boolean, optional
        verbose mode
    """
    folds = []
    keys = []
    if cfg["stratified"] and y.ndim > 1:
        raise TypeError("Stratified and multi-target are incompatible.")
    else:
        # construct the outer cross-validation folds iterator
        fg = _construct_folds_iterator(y, cfg, "crossval_score")
        for i, (train, test) in enumerate(fg):
            if verbose:
                print("Outer Fold {}:\n{}\n{}\n---".format(
                        i, np.sort(y[train]), np.sort(y[test])))
            folds.append((train, test))
            keys.append("%d" % i)
            if cfg["modelSelection"]:
                yi = y[train]
                # construct the inner cross-validation folds iterator
                fgi = _construct_folds_iterator(yi, cfg, "gridSearch")
                for k, (itrain, itest) in enumerate(fgi):
                    if verbose:
                        print("Inner Fold {}:\n{}\n{}\n---".format(
                                k, np.sort(yi[itrain]), np.sort(yi[itest])))
                    folds.append((itrain, itest))
                    keys.append("%d_%d" % (i, k))
    # associate string identifiers with corresponding folds
    if cfg["modelSelection"]:
        folds_dic = dict(zip(keys, folds), n_inner=(k + 1), n_outer=(i + 1))
    else:
        folds_dic = dict(zip(keys, folds), n_outer=(i + 1))
    return folds_dic


def get_fold(folds, outer, inner=None):
    """Get a given folds from the folds dict (see make_folds)

    Parameters
    ----------
    folds : dict,
        dictionnary with all the folds.
    outer : int,
        ID of the outer fold.
    inner : int, optional (default=None)
        ID of the inner fold.
    """
    try:
        if inner is None:
            return folds["%d" % outer]
        else:
            outer_ind = folds["%d" % outer][0]
            inner_ind = folds["%d_%d" % (outer, inner)]
            return outer_ind[inner_ind[0]], outer_ind[inner_ind[1]]
    except KeyError:
        raise KeyError("unexpected fold: outer={}, inner={}".format(
                outer, inner))


def print_fold(train_index, test_index):
    """Pretty printer for a couple of train/test folds.

    """
    print(train_index)
    print(test_index)
    print("train/test sizes: {}/{}".format(
            train_index.size,
            test_index.size))
    print("=======")



#!/usr/bin/env python
# Author: Benoit Da Mota <damota.benoit@gmail.com>
#
# License: BSD 3 clause
"""
Build commands and dependencies from JSON configuration files.
"""
import os
import json
import warnings

import numpy as np
import sklearn.externals.joblib as joblib

from mempamal.crossval import make_folds
from mempamal.configuration import check_conf
from mempamal.workflow import create_wf, save_wf
from mempamal.dynamic import load_data, get_grid
from mempamal.arguments import get_cmd_builder_argparser

verbose = False


if __name__ == "__main__":
    # parse command line arguments
    args = get_cmd_builder_argparser().parse_args()
    if args.no_warn:
        warnings.simplefilter("ignore")
    else:
        warnings.simplefilter("default")
    verbose = args.verbose
    if verbose:
        print("=======")
        print(args)
        print("=======")

    # step 1: read data file
    with open(args.data, 'r') as fd:
            data_cfg = dict(json.load(fd), src=args.data)
    check_conf(data_cfg, cat="data", verbose=verbose)
    X, Y = load_data(data_cfg)
    n_samples = X.shape[0]
    n_targets = 1 if (Y.ndim == 1) else Y.shape[1]
    if n_targets > 1:
        warnings.warn("More than one target. Unexpected results or crashes "
                      "may occur if your methods and/or metrics "
                      "cannot handle multiple target.", RuntimeWarning)

    # step 2: read crossval file
    with open(args.crossval, 'r') as fd:
            cv_cfg = dict(json.load(fd),
                          src=os.path.basename(args.crossval))
    check_conf(cv_cfg, cat="crossval", verbose=verbose)

    # step 3: load method configuration
    with open(args.method, 'r') as fd:
        method_cfg = dict(json.load(fd),
                          src=os.path.basename(args.method))
    check_conf(method_cfg, cat="method", verbose=verbose)

    # step 4: write dataset file(s)
    # generate folds/grid and write dataset
    output_file = os.path.join(args.outputdir, "dataset.joblib")
    folds = dict(make_folds(Y, cv_cfg, verbose=verbose),
                 src=os.path.basename(output_file))
    if cv_cfg["modelSelection"]:
        grid = get_grid(cv_cfg, X, Y)
    else:
        grid = np.asarray([method_cfg[method_cfg["est_param"]]])
    if verbose:
        print("Input dataset destination: {}".format(output_file))
    dataset = {"X": X, "Y": Y,
               "n_samples": n_samples, "n_targets": n_targets,
               "folds": folds, "grid": grid}
    joblib.dump(dataset, output_file, compress=1)

    # step 5: construct commands or workflow
    wf = create_wf(folds, cv_cfg, data_cfg, method_cfg, verbose=verbose)
    wf_out = os.path.join(args.outputdir, "workflow.json")
    save_wf(wf, wf_out, mode=args.output_mode)
    if verbose:
        print("Write workflow to: {}".format(wf_out))



# Author: Benoit Da Mota <damota.benoit@gmail.com>
#
# License: BSD 3 clause
"""
Workflow generation.
"""
import os.path as path
import numpy as np


def _create_generic(folds_dic, cv_cfg, data_cfg, method_cfg,
                    mapper="./scripts/mapper.py",
                    i_red="./scripts/inner_reducer.py",
                    o_red="./scripts/outer_reducer.py",
                    verbose=False):
    # retrieve I/O directory
    in_out_dir = data_cfg["in_out_dir"]

    # construct paths
    cv = path.join(in_out_dir, cv_cfg["src"])
    folds = path.join(in_out_dir, folds_dic["src"])
    method = path.join(in_out_dir, method_cfg["src"])
    # results filenames
    m_out = path.join(in_out_dir, "map_res_{outer}_{inner}.pkl")
    ri_out = path.join(in_out_dir, "red_res_{outer}.pkl")
    ro_out = path.join(in_out_dir, "final_res.pkl")

    # number of folds
    n_o = folds_dic["n_outer"]
    n_i = folds_dic["n_inner"] if cv_cfg["modelSelection"] else None

    # a workflow is a collection of commands and dependancies
    all_cmd = {}
    # dependancies are tuples (cmd_A, cmd_B) stored in a dict
    # with cmd_B that waits for cmd_A completion
    dependancies = []
    name_ired = "|--- Inner reduce outer={}"
    name_ored = "|- Final reduce"

    for i in xrange(n_o):
        cmd_mapper = ["python", mapper, cv, method, folds]
        name_cur_ired = name_ired.format(i)
        if cv_cfg["modelSelection"]:
            for k in xrange(n_i):
                cur_cmd = (cmd_mapper +
                           [m_out.format(inner=k, outer=i), repr(i),
                            "--inner", repr(k)])
                name = "|----- Map outer={} inner={}".format(i, k)
                all_cmd[name] = cur_cmd
                dependancies.append((name, name_cur_ired))
                if verbose:
                    print(" ".join(cur_cmd))

            cmd_i_red = ["python", i_red, cv, method, folds,
                         ri_out.format(outer=i),
                         m_out.format(outer=i, inner="{inner}"),
                         repr(i)]
            all_cmd[name_cur_ired] = cmd_i_red
            dependancies.append((name_cur_ired, name_ored))
            if verbose:
                print("\n{}\n".format(" ".join(cmd_i_red)))
        else:
            cur_cmd = cmd_mapper + [ri_out.format(outer=i), repr(i)]
            name = "|--- Map outer={}".format(i)
            all_cmd[name] = cur_cmd
            dependancies.append((name, name_ored))
            if verbose:
                print(" ".join(cur_cmd))
    cmd_o_red = ["python", o_red, ro_out, ri_out]
    all_cmd[name_ored] = cmd_o_red
    if verbose:
        print(" ".join(cmd_o_red))
    return all_cmd, dependancies


def create_wf(folds_dic, cv_cfg, data_cfg, method_cfg, verbose=False):
    c_map = method_cfg["mapper"]
    c_i_red = method_cfg["inner_reducer"]
    c_o_red = method_cfg["outer_reducer"]
    return _create_generic(folds_dic, cv_cfg, data_cfg, method_cfg,
                           mapper=c_map, i_red=c_i_red, o_red=c_o_red,
                           verbose=verbose)


def save_wf(wf, output_file, mode="soma-workflow"):
    cmd = wf[0]
    dep_orig = wf[1]
    if mode == "soma-workflow":
        from soma_workflow.client import Job, Workflow, Helper
        for k, v in cmd.iteritems():
            cmd[k] = Job(command=v, name=k)
        dep = [((cmd[a], cmd[b])) for a, b in dep_orig]
        jobs = np.asarray(cmd.values())[np.argsort(cmd.keys())]
        workflow = Workflow(jobs=jobs,
                            dependencies=dep)
        Helper.serialize(output_file, workflow)
    elif mode == "cmd-list":
        import json
        for k, v in cmd.iteritems():
            cmd[k] = " ".join(v)
        with open(output_file, 'w') as fd:
            json.dump(dict(cmd=cmd, dep=dep_orig), fd, indent=True)
    else:
        raise TypeError("Invalid workflow mode \'{}\'".format(mode))



#!/usr/bin/env python
# Author: Benoit Da Mota <damota.benoit@gmail.com>
#
# License: BSD 3 clause
import setuptools

setuptools.setup(
    name="mempamal",
    version="0.1.0",
    description="MEMPAMAL: Means for EMbarrassingly PArallel MAchine Learning",
    author="Benoit Da Mota",
    author_email="damota.benoit@gmail.com",
    license='BSD 3-clause',
    url='https://github.com/BenoitDamota/mempamal',
    packages=["mempamal", "mempamal.datasets", "mempamal.scripts"],
    long_description="""A set of python helpers to produce and run
    embarrassingly parallel machine learning workflows""",
    install_requires=['numpy', 'scikit-learn'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: Unix',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python',
        'Topic :: Software Development',
        'Topic :: Scientific/Engineering',
        'Topic :: Software Development :: Libraries :: Python Modules']
    )
